"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
// Opciones para el QUERY

// Clause Where

// Crea los querys
class QueryBuilder {
  _columns = '*';
  _whereClause = undefined;
  _whereArgs = undefined;
  _orderBy = undefined;
  _limit = undefined;
  _page = undefined;
  constructor(db, tableName) {
    this.db = db;
    this.tableName = tableName;
  }

  // SELECT
  static buildSelect(tableName, options) {
    const {
      where
    } = options;
    let sqlParts = ['SELECT', options.columns ?? "*", 'FROM', tableName];
    if (where && where.clause) {
      sqlParts.push("WHERE");
      sqlParts.push(where.clause);
    }
    if (options.order) {
      sqlParts.push("ORDER BY");
      sqlParts.push(options.order);
    }
    if (options.limit) {
      sqlParts.push("LIMIT");
      sqlParts.push(options.limit);
      if (options.page) {
        sqlParts.push("OFFSET");
        sqlParts.push(options.limit * (options.page - 1));
      }
    }
    return sqlParts.filter(p => p !== '').join(' ');
  }

  // Creates the "INSERT" sql statement
  static buildInsert(tableName, object) {
    const keys = Object.keys(object);
    const columns = keys.join(', ');
    const values = keys.map(() => '?').join(', ');
    return `INSERT INTO ${tableName} (${columns}) VALUES (${values});`;
  }

  // Creates the "INSERT" sql statement
  static buildInsertArray(tableName, keys, data) {
    if (keys.length === 0) {
      throw new Error('fields is empty');
    }
    if (data.length === 0) {
      throw new Error('data is empty');
    }
    const columns = keys.join(', ');
    const values = keys.map(() => '?').join(', ');
    const valuesArray = data.map(() => `(${values})`).join(', ');
    return `INSERT INTO ${tableName} (${columns}) VALUES ${valuesArray};`;
  }

  // Creates the "Update" sql statement
  static buildUpdate(tableName, object, whereClause) {
    // Extrae el valor de "id" y crea un nuevo objeto sin ese dato
    //const { id, ...props } = object
    const values = Object.keys(object).map(k => `${k} = ?`).join(', ');
    const wherePart = whereClause && whereClause.trim().length > 0 ? ` WHERE ${whereClause}` : '';
    return `UPDATE ${tableName} SET ${values}${wherePart};`;
  }

  // Creates the "DELETE" sql statement
  static buildDelete(tableName, whereClause) {
    const wherePart = whereClause && whereClause.trim().length > 0 ? ` WHERE ${whereClause}` : '';
    return `DELETE FROM ${tableName}${wherePart};`;
  }

  /**
   * Compila una clausula where
   * 
   * @param {any[]} array [1,3,4,2]
   * @returns (?, ?, ?, ?)
   */
  static whereClausePlaceholders(array) {
    var args = [];
    for (let i = 0; i < array.length; i++) {
      args.push("?");
    }
    return "(" + args.join(", ") + ")";
  }

  /**
   * @param sqlString "valor con 'comillas'"
   * @returns "'valor con \'comillas\''"
   */
  static escapeSqlString(sqlString) {
    let escape = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '\'';
    var sb = [];
    sb.push(escape);
    if (sqlString.indexOf(escape) != -1) {
      var length = sqlString.length;
      for (var i = 0; i < length; i++) {
        var c = sqlString.charAt(i);
        if (c == escape) {
          sb.push(escape);
        }
        sb.push(c);
      }
    } else {
      sb.push(sqlString);
    }
    sb.push(escape);
    return sb.join("");
  }
  select(columns) {
    this._columns = columns;
    return this;
  }
  where(clause, args) {
    this._whereClause = clause;
    this._whereArgs = args;
    return this;
  }
  orderBy(orderBy) {
    this._orderBy = orderBy;
    return this;
  }
  limit(limit) {
    this._limit = limit;
    return this;
  }
  page(page) {
    this._page = page;
    return this;
  }

  /** 
   * SELECT
   * 
   * @returns array de registros
   */
  get() {
    const sql = QueryBuilder.buildSelect(this.tableName, {
      columns: this._columns,
      where: {
        clause: this._whereClause,
        args: this._whereArgs
      },
      order: this._orderBy,
      limit: this._limit,
      page: this._page
    });
    return this.db.executeSql(sql, this._whereArgs).then(result => result.rows);
  }

  /** 
   * INSERT
   * 
   * @param {any} values
   * @returns insertId
   */
  insert(values) {
    var sql = QueryBuilder.buildInsert(this.tableName, values);
    const params = Object.values(values);
    return this.db.executeSql(sql, params).then(result => result.insertId ?? -1);
  }

  /** 
   * INSERT
   * 
   * @param {string[]} fields ["Column 1", "Column 2"]
   * @param {any[][]} data [
   *  ["foo", "bar"],
   *  ["abc", "def"]
   * ]
   * @returns result
   */
  insertArray(fields, data) {
    const sql = QueryBuilder.buildInsertArray(this.tableName, fields, data);
    // crea una nueva matriz con todos los elementos de las submatrices 
    // concatenados recursivamente hasta la profundidad especificada
    const params = data.flat();
    return this.db.executeSql(sql, params);
  }

  /** 
   * UPDATE
   * 
   * @param {any} values
   * @returns rowsAffected
   */
  update(values) {
    const sql = QueryBuilder.buildUpdate(this.tableName, values, this._whereClause);
    const params = Object.values(values);
    const whereArgs = this._whereArgs ?? [];
    return this.db.executeSql(sql, [...params, ...whereArgs]).then(result => result.rowsAffected);
  }

  /** 
   * DELETE
   * 
   * @returns rowsAffected
   */
  delete() {
    const sql = QueryBuilder.buildDelete(this.tableName, this._whereClause);
    return this.db.executeSql(sql, this._whereArgs).then(result => result.rowsAffected);
  }
}
var _default = exports.default = QueryBuilder;
//# sourceMappingURL=QueryBuilder.js.map